"""Monkey-patch for the `httpx` library (sync and async)."""

import logging
import time
from urllib.parse import urlparse

from smello.capture import serialize_request_response
from smello.config import SmelloConfig
from smello.transport import send

logger = logging.getLogger(__name__)


def patch_httpx(config: SmelloConfig) -> None:
    """Patch httpx.Client.send and httpx.AsyncClient.send."""
    try:
        import httpx  # noqa: PLC0415 -- optional dependency
    except ImportError:
        return  # httpx not installed, skip

    _patch_sync(httpx, config)
    _patch_async(httpx, config)


def _patch_sync(httpx, config: SmelloConfig) -> None:
    original_send = httpx.Client.send

    def patched_send(self, request, **kwargs):
        host = urlparse(str(request.url)).hostname or ""

        if not config.should_capture(host):
            return original_send(self, request, **kwargs)

        start = time.monotonic()
        response = original_send(self, request, **kwargs)
        duration = time.monotonic() - start

        try:
            payload = serialize_request_response(
                config=config,
                method=request.method,
                url=str(request.url),
                request_headers=dict(request.headers),
                request_body=request.content,
                status_code=response.status_code,
                response_headers=dict(response.headers),
                response_body=response.text,
                duration_s=duration,
                library="httpx",
            )
            send(payload)
        except Exception as err:
            logger.debug("Failed to capture request: %s", err)

        return response

    httpx.Client.send = patched_send


def _patch_async(httpx, config: SmelloConfig) -> None:
    original_send = httpx.AsyncClient.send

    async def patched_send(self, request, **kwargs):
        host = urlparse(str(request.url)).hostname or ""

        if not config.should_capture(host):
            return await original_send(self, request, **kwargs)

        start = time.monotonic()
        response = await original_send(self, request, **kwargs)
        duration = time.monotonic() - start

        try:
            payload = serialize_request_response(
                config=config,
                method=request.method,
                url=str(request.url),
                request_headers=dict(request.headers),
                request_body=request.content,
                status_code=response.status_code,
                response_headers=dict(response.headers),
                response_body=response.text,
                duration_s=duration,
                library="httpx",
            )
            send(payload)
        except Exception as err:
            logger.debug("Failed to capture request: %s", err)

        return response

    httpx.AsyncClient.send = patched_send



"""Configuration for Smello client."""

from dataclasses import dataclass, field


@dataclass
class SmelloConfig:
    server_url: str
    capture_hosts: list[str] = field(default_factory=list)
    capture_all: bool = True
    ignore_hosts: list[str] = field(default_factory=list)
    redact_headers: list[str] = field(
        default_factory=lambda: ["authorization", "x-api-key"]
    )
    redact_query_params: list[str] = field(default_factory=list)

    def should_capture(self, host: str) -> bool:
        """Decide whether to capture a request to the given host."""
        if host in self.ignore_hosts:
            return False
        if self.capture_all:
            return True
        return host in self.capture_hosts



"""Tests for smello.config.SmelloConfig."""

import pytest
from smello.config import SmelloConfig


@pytest.fixture()
def default_config():
    return SmelloConfig(server_url="http://test:5110")


@pytest.fixture()
def selective_config():
    return SmelloConfig(
        server_url="http://test:5110",
        capture_all=False,
        capture_hosts=["api.stripe.com"],
    )


def test_capture_all_by_default(default_config):
    assert default_config.should_capture("api.stripe.com") is True
    assert default_config.should_capture("example.com") is True


def test_ignore_hosts():
    config = SmelloConfig(
        server_url="http://test:5110", ignore_hosts=["localhost", "127.0.0.1"]
    )
    assert config.should_capture("localhost") is False
    assert config.should_capture("127.0.0.1") is False
    assert config.should_capture("api.stripe.com") is True


def test_capture_specific_hosts_only(selective_config):
    assert selective_config.should_capture("api.stripe.com") is True
    assert selective_config.should_capture("api.openai.com") is False


def test_ignore_takes_precedence_over_capture_hosts():
    config = SmelloConfig(
        server_url="http://test:5110",
        capture_all=False,
        capture_hosts=["api.stripe.com"],
        ignore_hosts=["api.stripe.com"],
    )
    assert config.should_capture("api.stripe.com") is False


def test_ignore_takes_precedence_over_capture_all():
    config = SmelloConfig(
        server_url="http://test:5110",
        capture_all=True,
        ignore_hosts=["secret.internal"],
    )
    assert config.should_capture("secret.internal") is False
    assert config.should_capture("anything.else") is True



"""Tests for smello.transport."""

import json
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer

import pytest
from smello.transport import _json_default, flush, send, shutdown, start_worker


class _CaptureHandler(BaseHTTPRequestHandler):
    captured: list = []

    def do_POST(self):
        length = int(self.headers.get("Content-Length", 0))
        body = json.loads(self.rfile.read(length))
        _CaptureHandler.captured.append(body)
        self.send_response(201)
        self.end_headers()
        self.wfile.write(b'{"status":"ok"}')

    def log_message(self, format, *args):
        pass


@pytest.fixture()
def capture_server():
    """Start a minimal HTTP server that records POSTed payloads."""
    _CaptureHandler.captured = []
    server = HTTPServer(("127.0.0.1", 0), _CaptureHandler)
    port = server.server_address[1]
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    yield f"http://127.0.0.1:{port}", _CaptureHandler.captured
    server.shutdown()


def test_send_delivers_payload(capture_server):
    url, captured = capture_server
    start_worker(url)

    payload = {
        "id": "test-transport-1",
        "request": {"method": "GET", "url": "https://example.com"},
        "response": {"status_code": 200},
    }
    send(payload)

    deadline = time.monotonic() + 5
    while not captured and time.monotonic() < deadline:
        time.sleep(0.05)

    assert len(captured) == 1
    assert captured[0]["id"] == "test-transport-1"


def test_flush_waits_for_pending_payloads(capture_server):
    url, captured = capture_server
    start_worker(url)

    for i in range(5):
        send({"id": f"flush-{i}", "request": {}, "response": {}})

    result = flush(timeout=5.0)

    assert result is True
    assert len(captured) == 5


def test_flush_returns_true_when_queue_already_empty(capture_server):
    url, _captured = capture_server
    start_worker(url)

    result = flush(timeout=1.0)
    assert result is True


def test_shutdown_flushes(capture_server):
    url, captured = capture_server
    start_worker(url)

    send({"id": "shutdown-1", "request": {}, "response": {}})

    result = shutdown(timeout=5.0)

    assert result is True
    assert len(captured) == 1


# ---------------------------------------------------------------------------
# _json_default() — fallback serializer
# ---------------------------------------------------------------------------


def test_json_default_handles_bytes():
    assert _json_default(b"\x00\x01\x02") == "b'\\x00\\x01\\x02'"


def test_json_default_handles_arbitrary_objects():
    class Custom:
        def __repr__(self):
            return "Custom()"

    assert _json_default(Custom()) == "Custom()"


def test_json_default_survives_broken_repr():
    class Broken:
        def __repr__(self):
            raise RuntimeError("boom")

    assert _json_default(Broken()) == "<unserializable>"


def test_json_dumps_with_default_serializes_bytes():
    """bytes in a payload dict should serialize via the fallback."""
    payload = {"headers": {"x-bin": b"\n\x02"}, "body": "ok"}
    result = json.loads(json.dumps(payload, default=_json_default))
    assert isinstance(result["headers"]["x-bin"], str)


def test_send_delivers_payload_with_bytes(capture_server):
    """Payloads containing bytes should not crash the transport."""
    url, captured = capture_server
    start_worker(url)

    send({"id": "bytes-test", "headers": {"bin": b"\xff"}, "response": {}})

    deadline = time.monotonic() + 5
    while not captured and time.monotonic() < deadline:
        time.sleep(0.05)

    assert len(captured) == 1
    assert captured[0]["id"] == "bytes-test"



"""Background transport: sends captured data to the Smello server without blocking."""

import json
import logging
import queue
import threading
import urllib.request

logger = logging.getLogger(__name__)

_queue: queue.Queue = queue.Queue(maxsize=1000)
_server_url: str = ""
_started: bool = False


def start_worker(server_url: str) -> None:
    """Start the background worker thread."""
    global _server_url, _started
    _server_url = server_url

    if _started:
        return
    _started = True

    thread = threading.Thread(target=_worker, daemon=True, name="smello-transport")
    thread.start()


def send(payload: dict) -> None:
    """Queue a capture payload for sending. Non-blocking, drops if queue is full."""
    try:
        _queue.put_nowait(payload)
    except queue.Full:
        logger.warning("Payload dropped: capture queue is full")


def flush(timeout: float = 2.0) -> bool:
    """Block until all queued payloads are sent, or *timeout* seconds elapse.

    Returns ``True`` if the queue drained in time, ``False`` otherwise.
    """
    # Queue.join() has no timeout parameter. Access the underlying
    # condition variable directly — same technique Sentry's SDK uses.
    with _queue.all_tasks_done:
        if _queue.unfinished_tasks:
            logger.debug("Flushing %d pending capture(s)…", _queue.unfinished_tasks)
            _queue.all_tasks_done.wait(timeout=timeout)

    drained = _queue.unfinished_tasks == 0
    if not drained:
        logger.warning(
            "Flush timed out with %d capture(s) still pending",
            _queue.unfinished_tasks,
        )
    return drained


def shutdown(timeout: float = 2.0) -> bool:
    """Flush pending payloads then stop accepting new ones.

    Returns ``True`` if the queue drained in time, ``False`` otherwise.
    """
    result = flush(timeout=timeout)
    return result


def _worker() -> None:
    """Background worker that sends queued payloads to the server."""
    while True:
        payload = _queue.get()
        try:
            _send_to_server(payload)
        except Exception as err:
            logger.warning("Failed to send capture to %s: %s", _server_url, err)
        _queue.task_done()


def _json_default(obj: object) -> str:
    """Fallback serializer for types that json.dumps cannot handle (e.g. bytes)."""
    try:
        return repr(obj)
    except Exception:
        return "<unserializable>"


def _send_to_server(payload: dict) -> None:
    """Send a payload to the Smello server using urllib (to avoid recursion)."""
    data = json.dumps(payload, default=_json_default).encode("utf-8")
    req = urllib.request.Request(
        f"{_server_url}/api/capture",
        data=data,
        headers={"Content-Type": "application/json"},
        method="POST",
    )
    urllib.request.urlopen(req, timeout=5)



"""Tests for smello.patches.patch_grpc — gRPC interception."""

import sys
import types
from unittest.mock import MagicMock, patch

import pytest
from smello.config import SmelloConfig
from smello.patches.patch_grpc import (
    _GRPC_STATUS_TO_HTTP,
    _extract_host,
    _grpc_status_to_http,
    _intercept_unary_unary,
    _make_interceptor_class,
    _metadata_to_dict,
    _proto_to_json,
    _send_capture,
    patch_grpc,
)


@pytest.fixture()
def config():
    return SmelloConfig(
        server_url="http://test:5110", redact_headers=["authorization", "x-api-key"]
    )


# ---------------------------------------------------------------------------
# _make_mock_grpc helper
# ---------------------------------------------------------------------------


def _make_mock_grpc():
    """Create a mock grpc module with the needed attributes."""
    mock_grpc = types.ModuleType("grpc")
    mock_grpc.insecure_channel = MagicMock(name="insecure_channel")
    mock_grpc.secure_channel = MagicMock(name="secure_channel")
    mock_grpc.intercept_channel = MagicMock(name="intercept_channel")
    mock_grpc.UnaryUnaryClientInterceptor = type("UnaryUnaryClientInterceptor", (), {})
    return mock_grpc


# ---------------------------------------------------------------------------
# patch_grpc()
# ---------------------------------------------------------------------------


def test_patch_skips_when_grpc_not_installed(config):
    """patch_grpc() should return silently when grpc is not importable."""
    with patch.dict(sys.modules, {"grpc": None}):
        patch_grpc(config)


def test_insecure_channel_wraps_with_interceptor(config):
    mock_grpc = _make_mock_grpc()
    original_insecure = mock_grpc.insecure_channel
    mock_channel = MagicMock()
    original_insecure.return_value = mock_channel

    with patch.dict(sys.modules, {"grpc": mock_grpc}):
        patch_grpc(config)

    mock_grpc.insecure_channel("localhost:50051")

    original_insecure.assert_called_once_with(
        "localhost:50051", options=None, compression=None
    )
    mock_grpc.intercept_channel.assert_called_once()
    args = mock_grpc.intercept_channel.call_args
    assert args[0][0] is mock_channel


def test_secure_channel_wraps_with_interceptor(config):
    mock_grpc = _make_mock_grpc()
    original_secure = mock_grpc.secure_channel
    mock_channel = MagicMock()
    original_secure.return_value = mock_channel
    mock_creds = MagicMock()

    with patch.dict(sys.modules, {"grpc": mock_grpc}):
        patch_grpc(config)

    mock_grpc.secure_channel("host:443", mock_creds)

    original_secure.assert_called_once_with(
        "host:443", mock_creds, options=None, compression=None
    )
    mock_grpc.intercept_channel.assert_called_once()
    args = mock_grpc.intercept_channel.call_args
    assert args[0][0] is mock_channel


def test_secure_channel_forwards_options_and_compression(config):
    """Extra kwargs like options should be forwarded to the original."""
    mock_grpc = _make_mock_grpc()
    original_secure = mock_grpc.secure_channel
    original_secure.return_value = MagicMock()
    mock_creds = MagicMock()

    with patch.dict(sys.modules, {"grpc": mock_grpc}):
        patch_grpc(config)

    opts = [("grpc.max_send_message_length", -1)]
    mock_grpc.secure_channel("host:443", mock_creds, options=opts, compression=1)

    original_secure.assert_called_once_with(
        "host:443", mock_creds, options=opts, compression=1
    )


# ---------------------------------------------------------------------------
# _make_interceptor_class()
# ---------------------------------------------------------------------------


def test_make_interceptor_class_inherits_base():
    """Returned class should inherit from the provided base class."""
    base = type("FakeBase", (), {})
    cls = _make_interceptor_class(base)
    assert issubclass(cls, base)


def test_make_interceptor_class_delegates_to_intercept(config):
    """intercept_unary_unary should delegate to _intercept_unary_unary."""
    base = type("FakeBase", (), {})
    cls = _make_interceptor_class(base)
    interceptor = cls(config, "host:443")

    with patch("smello.patches.patch_grpc._intercept_unary_unary") as mock_fn:
        mock_fn.return_value = "result"
        result = interceptor.intercept_unary_unary("cont", "details", "req")

    assert result == "result"
    mock_fn.assert_called_once_with(config, "host:443", "cont", "details", "req")


# ---------------------------------------------------------------------------
# _intercept_unary_unary() — success path
# ---------------------------------------------------------------------------


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_interceptor_captures_successful_call(mock_serialize, mock_send, config):
    call_details = MagicMock()
    call_details.method = "/pkg.Service/Method"
    call_details.metadata = [("x-custom", "val")]

    mock_request = MagicMock()
    mock_request.__str__ = lambda self: "request_str"

    mock_response = MagicMock()
    mock_result = MagicMock()
    mock_result.__str__ = lambda self: "result_str"
    mock_response.result.return_value = mock_result
    mock_response.trailing_metadata.return_value = []

    continuation = MagicMock(return_value=mock_response)
    mock_serialize.return_value = {"id": "test"}

    result = _intercept_unary_unary(
        config, "api.example.com:443", continuation, call_details, mock_request
    )

    assert result is mock_response
    continuation.assert_called_once_with(call_details, mock_request)
    mock_serialize.assert_called_once()

    kw = mock_serialize.call_args[1]
    assert kw["method"] == "POST"
    assert kw["url"] == "grpc://api.example.com:443/pkg.Service/Method"
    assert kw["status_code"] == 200
    assert kw["library"] == "grpc"
    assert kw["request_headers"] == {"x-custom": "val"}

    mock_send.assert_called_once_with({"id": "test"})


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_trailing_metadata_merged_into_response_headers(
    mock_serialize, mock_send, config
):
    """Trailing metadata from the response should appear in response_headers."""
    call_details = MagicMock()
    call_details.method = "/pkg.Service/Method"
    call_details.metadata = None

    mock_response = MagicMock()
    mock_response.result.return_value = MagicMock()
    mock_response.trailing_metadata.return_value = [
        ("x-trace-id", "abc123"),
        ("x-request-id", "def456"),
    ]

    continuation = MagicMock(return_value=mock_response)
    mock_serialize.return_value = {"id": "test"}

    _intercept_unary_unary(config, "host:443", continuation, call_details, MagicMock())

    kw = mock_serialize.call_args[1]
    assert kw["response_headers"]["grpc-status"] == "0"
    assert kw["response_headers"]["grpc-status-name"] == "OK"
    assert kw["response_headers"]["x-trace-id"] == "abc123"
    assert kw["response_headers"]["x-request-id"] == "def456"


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_success_still_returned_when_send_capture_fails(
    mock_serialize, mock_send, config
):
    """If _send_capture raises, the response should still be returned."""
    call_details = MagicMock()
    call_details.method = "/svc/Method"
    call_details.metadata = None

    mock_response = MagicMock()
    mock_response.result.return_value = MagicMock()
    mock_response.trailing_metadata.return_value = []
    continuation = MagicMock(return_value=mock_response)

    mock_serialize.side_effect = RuntimeError("serialize boom")

    result = _intercept_unary_unary(
        config, "host:443", continuation, call_details, MagicMock()
    )

    assert result is mock_response


# ---------------------------------------------------------------------------
# _intercept_unary_unary() — error path
# ---------------------------------------------------------------------------


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_interceptor_captures_error_call(mock_serialize, mock_send, config):
    call_details = MagicMock()
    call_details.method = "/pkg.Service/Method"
    call_details.metadata = None
    mock_request = MagicMock()

    mock_code = MagicMock()
    mock_code.value = (5,)  # NOT_FOUND
    mock_code.name = "NOT_FOUND"

    rpc_error = Exception("not found")
    rpc_error.code = MagicMock(return_value=mock_code)

    mock_response = MagicMock()
    mock_response.result.side_effect = rpc_error

    continuation = MagicMock(return_value=mock_response)
    mock_serialize.return_value = {"id": "err"}

    with pytest.raises(Exception, match="not found"):
        _intercept_unary_unary(
            config, "api.example.com:443", continuation, call_details, mock_request
        )

    mock_serialize.assert_called_once()
    kw = mock_serialize.call_args[1]
    assert kw["status_code"] == 404
    assert kw["response_headers"]["grpc-status"] == "5"
    assert kw["response_headers"]["grpc-status-name"] == "NOT_FOUND"
    mock_send.assert_called_once_with({"id": "err"})


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_error_without_code_attr_falls_back_to_unknown(
    mock_serialize, mock_send, config
):
    """A plain Exception (no .code()) should map to UNKNOWN / 500."""
    call_details = MagicMock()
    call_details.method = "/svc/Method"
    call_details.metadata = None

    plain_error = ValueError("something broke")

    mock_response = MagicMock()
    mock_response.result.side_effect = plain_error
    continuation = MagicMock(return_value=mock_response)
    mock_serialize.return_value = {"id": "unk"}

    with pytest.raises(ValueError, match="something broke"):
        _intercept_unary_unary(
            config, "host:443", continuation, call_details, MagicMock()
        )

    kw = mock_serialize.call_args[1]
    assert kw["status_code"] == 500
    assert kw["response_headers"]["grpc-status"] == "2"
    assert kw["response_headers"]["grpc-status-name"] == "UNKNOWN"
    assert "something broke" in kw["response_body"]


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_error_still_raised_when_send_capture_fails(mock_serialize, mock_send, config):
    """If _send_capture raises on the error path, the original error is still re-raised."""
    call_details = MagicMock()
    call_details.method = "/svc/Method"
    call_details.metadata = None

    rpc_error = Exception("original error")
    mock_response = MagicMock()
    mock_response.result.side_effect = rpc_error
    continuation = MagicMock(return_value=mock_response)

    mock_serialize.side_effect = RuntimeError("capture boom")

    with pytest.raises(Exception, match="original error"):
        _intercept_unary_unary(
            config, "host:443", continuation, call_details, MagicMock()
        )


# ---------------------------------------------------------------------------
# _intercept_unary_unary() — skip path
# ---------------------------------------------------------------------------


@patch("smello.patches.patch_grpc.serialize_request_response")
def test_interceptor_skips_ignored_host(mock_serialize, config):
    config.ignore_hosts = ["ignored.example.com"]

    call_details = MagicMock()
    call_details.method = "/pkg.Service/Method"
    mock_request = MagicMock()
    mock_response = MagicMock()
    continuation = MagicMock(return_value=mock_response)

    result = _intercept_unary_unary(
        config, "ignored.example.com:443", continuation, call_details, mock_request
    )

    assert result is mock_response
    mock_serialize.assert_not_called()


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_bytes_method_decoded(mock_serialize, mock_send, config):
    call_details = MagicMock()
    call_details.method = b"/pkg.Service/Method"
    call_details.metadata = None

    mock_response = MagicMock()
    mock_response.result.return_value = MagicMock()
    mock_response.trailing_metadata.return_value = []
    continuation = MagicMock(return_value=mock_response)
    mock_serialize.return_value = {"id": "test"}

    _intercept_unary_unary(config, "host:443", continuation, call_details, MagicMock())

    kw = mock_serialize.call_args[1]
    assert kw["url"] == "grpc://host:443/pkg.Service/Method"


# ---------------------------------------------------------------------------
# _send_capture()
# ---------------------------------------------------------------------------


@patch("smello.patches.patch_grpc.send")
@patch("smello.patches.patch_grpc.serialize_request_response")
def test_send_capture(mock_serialize, mock_send, config):
    mock_serialize.return_value = {"id": "payload"}

    _send_capture(
        config=config,
        method="POST",
        url="grpc://host:443/svc/Method",
        request_headers={"k": "v"},
        request_body='{"a": 1}',
        status_code=200,
        response_headers={"grpc-status": "0"},
        response_body='{"b": 2}',
        duration_s=0.5,
    )

    mock_serialize.assert_called_once_with(
        config=config,
        method="POST",
        url="grpc://host:443/svc/Method",
        request_headers={"k": "v"},
        request_body='{"a": 1}',
        status_code=200,
        response_headers={"grpc-status": "0"},
        response_body='{"b": 2}',
        duration_s=0.5,
        library="grpc",
    )
    mock_send.assert_called_once_with({"id": "payload"})


# ---------------------------------------------------------------------------
# _grpc_status_to_http()
# ---------------------------------------------------------------------------


@pytest.mark.parametrize(
    ("grpc_code", "expected_http"),
    list(_GRPC_STATUS_TO_HTTP.items()),
)
def test_grpc_status_to_http_mapping(grpc_code, expected_http):
    assert _grpc_status_to_http(grpc_code) == expected_http


def test_grpc_status_to_http_unknown_code():
    assert _grpc_status_to_http(99) == 500


# ---------------------------------------------------------------------------
# _extract_host()
# ---------------------------------------------------------------------------


@pytest.mark.parametrize(
    ("target", "expected"),
    [
        ("host:443", "host"),
        ("dns:///host:443", "host"),
        ("dns:///host", "host"),
        ("host", "host"),
        ("dns://authority/host:443", "host"),
        ("ipv4:127.0.0.1:50051", "127.0.0.1"),
    ],
)
def test_extract_host_variants(target, expected):
    assert _extract_host(target) == expected


# ---------------------------------------------------------------------------
# _proto_to_json()
# ---------------------------------------------------------------------------


def test_proto_to_json_fallback():
    obj = MagicMock()
    obj.__str__ = lambda self: "mock_string_repr"

    with patch.dict(sys.modules, {"google.protobuf.json_format": None}):
        result = _proto_to_json(obj)

    assert result == "mock_string_repr"


def test_proto_to_json_with_protobuf():
    mock_json_format = MagicMock()
    mock_json_format.MessageToJson.return_value = '{"field": "value"}'

    with patch.dict(
        sys.modules,
        {
            "google": MagicMock(),
            "google.protobuf": MagicMock(),
            "google.protobuf.json_format": mock_json_format,
        },
    ):
        msg = MagicMock()
        del msg._pb  # ensure no _pb attr — raw protobuf
        result = _proto_to_json(msg)

    assert result == '{"field": "value"}'
    mock_json_format.MessageToJson.assert_called_once_with(msg)


def test_proto_to_json_unwraps_proto_plus():
    """proto-plus wrappers (Google Cloud libraries) should be unwrapped via _pb."""
    mock_json_format = MagicMock()
    mock_json_format.MessageToJson.return_value = '{"unwrapped": true}'

    with patch.dict(
        sys.modules,
        {
            "google": MagicMock(),
            "google.protobuf": MagicMock(),
            "google.protobuf.json_format": mock_json_format,
        },
    ):
        inner_pb = MagicMock()
        wrapper = MagicMock()
        wrapper._pb = inner_pb
        result = _proto_to_json(wrapper)

    assert result == '{"unwrapped": true}'
    mock_json_format.MessageToJson.assert_called_once_with(inner_pb)


# ---------------------------------------------------------------------------
# _metadata_to_dict()
# ---------------------------------------------------------------------------


def test_metadata_to_dict_none():
    assert _metadata_to_dict(None) == {}


def test_metadata_to_dict_tuples():
    metadata = [("key1", "val1"), ("key2", "val2")]
    assert _metadata_to_dict(metadata) == {"key1": "val1", "key2": "val2"}


def test_metadata_to_dict_preserves_bytes():
    """Binary metadata values should pass through (transport handles serialization)."""
    metadata = [
        ("x-trace-id", "abc123"),
        ("pc-low-bwd-bin", b"\n\x02 \x10"),
    ]
    result = _metadata_to_dict(metadata)
    assert result["x-trace-id"] == "abc123"
    assert result["pc-low-bwd-bin"] == b"\n\x02 \x10"



"""Read SMELLO_* environment variables."""

from __future__ import annotations

import os

_TRUTHY = frozenset({"true", "1", "yes"})
_FALSY = frozenset({"false", "0", "no"})


def _env_str(name: str) -> str | None:
    """Read ``SMELLO_{name}`` from the environment.

    Returns the stripped value, or ``None`` if unset or empty.
    """
    value = os.environ.get(f"SMELLO_{name}", "").strip()
    return value if value else None


def _env_bool(name: str) -> bool | None:
    """Read ``SMELLO_{name}`` as a boolean.

    Truthy: ``true``, ``1``, ``yes`` (case-insensitive).
    Falsy:  ``false``, ``0``, ``no`` (case-insensitive).
    Returns ``None`` if unset, empty, or unrecognised.
    """
    raw = _env_str(name)
    if raw is None:
        return None
    lowered = raw.lower()
    if lowered in _TRUTHY:
        return True
    if lowered in _FALSY:
        return False
    return None


def _env_list(name: str) -> list[str] | None:
    """Read ``SMELLO_{name}`` as a comma-separated list.

    Returns a list of stripped, non-empty items, or ``None`` if the
    variable is unset or empty.
    """
    raw = _env_str(name)
    if raw is None:
        return None
    items = [item.strip() for item in raw.split(",") if item.strip()]
    return items if items else None



"""Monkey-patches for HTTP client libraries."""

from smello.config import SmelloConfig

# Alias imports to avoid shadowing the submodule names (patch_grpc,
# patch_httpx, patch_requests).  unittest.mock on Python < 3.12 resolves
# "smello.patches.patch_grpc.send" via getattr, so the module attribute
# must remain the *module*, not the function.
from smello.patches.patch_grpc import patch_grpc as _patch_grpc
from smello.patches.patch_httpx import patch_httpx as _patch_httpx
from smello.patches.patch_requests import patch_requests as _patch_requests


def apply_all(config: SmelloConfig) -> None:
    """Apply all available patches."""
    _patch_requests(config)
    _patch_httpx(config)
    _patch_grpc(config)



"""Monkey-patch for the `grpc` library (unary-unary calls)."""

import logging
import time

from smello.capture import serialize_request_response
from smello.config import SmelloConfig
from smello.transport import send

logger = logging.getLogger(__name__)

_GRPC_STATUS_TO_HTTP = {
    0: 200,  # OK
    1: 499,  # CANCELLED
    2: 500,  # UNKNOWN
    3: 400,  # INVALID_ARGUMENT
    4: 504,  # DEADLINE_EXCEEDED
    5: 404,  # NOT_FOUND
    6: 409,  # ALREADY_EXISTS
    7: 403,  # PERMISSION_DENIED
    8: 429,  # RESOURCE_EXHAUSTED
    9: 400,  # FAILED_PRECONDITION
    10: 409,  # ABORTED
    11: 400,  # OUT_OF_RANGE
    12: 501,  # UNIMPLEMENTED
    13: 500,  # INTERNAL
    14: 503,  # UNAVAILABLE
    15: 500,  # DATA_LOSS
    16: 401,  # UNAUTHENTICATED
}


def patch_grpc(config: SmelloConfig) -> None:
    """Patch grpc.insecure_channel and grpc.secure_channel to capture unary-unary calls."""
    try:
        import grpc  # type: ignore[unresolved-import]  # noqa: PLC0415 -- optional dependency
    except ImportError:
        return  # grpc not installed, skip

    # Define interceptor class here so it can inherit from the real
    # grpc.UnaryUnaryClientInterceptor (required by grpc.intercept_channel).
    Interceptor = _make_interceptor_class(grpc.UnaryUnaryClientInterceptor)

    original_insecure = grpc.insecure_channel
    original_secure = grpc.secure_channel

    def patched_insecure_channel(target, options=None, compression=None):
        channel = original_insecure(target, options=options, compression=compression)
        return grpc.intercept_channel(channel, Interceptor(config, target))

    def patched_secure_channel(target, credentials, options=None, compression=None):
        channel = original_secure(
            target, credentials, options=options, compression=compression
        )
        return grpc.intercept_channel(channel, Interceptor(config, target))

    grpc.insecure_channel = patched_insecure_channel
    grpc.secure_channel = patched_secure_channel


def _grpc_status_to_http(code_value: int) -> int:
    return _GRPC_STATUS_TO_HTTP.get(code_value, 500)


def _extract_host(target: str) -> str:
    """Extract hostname from a gRPC target string, stripping prefixes and port."""
    # Strip common gRPC target prefixes
    for prefix in ("dns:///", "dns://", "ipv4:", "ipv6:", "unix:"):
        if target.startswith(prefix):
            target = target[len(prefix) :]
            break

    # dns://authority/host — strip the authority component
    if "/" in target:
        target = target.rsplit("/", 1)[-1]

    # Remove port
    if ":" in target:
        target = target.rsplit(":", 1)[0]

    return target


def _metadata_to_dict(metadata) -> dict:
    """Convert gRPC metadata (list of tuples) to a dict."""
    if metadata is None:
        return {}
    return {k: v for k, v in metadata}


def _proto_to_json(message) -> str:
    """Convert a protobuf message to JSON string, falling back to str().

    Handles both raw protobuf messages and proto-plus wrappers (used by
    Google Cloud client libraries).  Proto-plus objects expose the
    underlying protobuf via a ``_pb`` attribute.
    """
    try:
        from google.protobuf.json_format import (  # type: ignore[unresolved-import]  # noqa: PLC0415 -- optional dependency
            MessageToJson,
        )

        # proto-plus wraps protobuf messages; unwrap before serializing.
        pb = getattr(message, "_pb", message)
        return MessageToJson(pb)
    except Exception:
        return str(message)


def _make_interceptor_class(base_class):
    """Create the interceptor class with the correct gRPC base class.

    We can't inherit from grpc.UnaryUnaryClientInterceptor at module level
    because grpc is an optional dependency. This factory is called from
    patch_grpc() after grpc has been successfully imported.
    """

    class _SmelloInterceptor(base_class):
        """gRPC unary-unary client interceptor that captures calls."""

        def __init__(self, config: SmelloConfig, target: str):
            self._config = config
            self._target = target

        def intercept_unary_unary(self, continuation, client_call_details, request):
            return _intercept_unary_unary(
                self._config, self._target, continuation, client_call_details, request
            )

    return _SmelloInterceptor


def _intercept_unary_unary(config, target, continuation, client_call_details, request):
    host = _extract_host(target)

    if not config.should_capture(host):
        return continuation(client_call_details, request)

    method = client_call_details.method
    if isinstance(method, bytes):
        method = method.decode("utf-8")

    url = f"grpc://{target}{method}"

    request_headers = _metadata_to_dict(client_call_details.metadata)
    request_body = _proto_to_json(request)

    start = time.monotonic()
    try:
        response = continuation(client_call_details, request)
        result = response.result()
        duration = time.monotonic() - start

        status_code = _grpc_status_to_http(0)
        response_body = _proto_to_json(result)
        response_headers = {
            "grpc-status": "0",
            "grpc-status-name": "OK",
        }
        trailing = getattr(response, "trailing_metadata", None)
        if callable(trailing):
            trailing = trailing()
        if trailing:
            response_headers.update(_metadata_to_dict(trailing))
    except Exception as err:
        duration = time.monotonic() - start

        grpc_code = 2  # UNKNOWN
        grpc_name = "UNKNOWN"
        response_body = str(err)

        if hasattr(err, "code") and callable(err.code):
            code_obj = err.code()  # type: ignore[call-top-callable]
            if hasattr(code_obj, "value"):
                grpc_code = code_obj.value[0]  # type: ignore[not-subscriptable]
                grpc_name = code_obj.name if hasattr(code_obj, "name") else grpc_name

        status_code = _grpc_status_to_http(grpc_code)
        response_headers = {
            "grpc-status": str(grpc_code),
            "grpc-status-name": grpc_name,
        }

        try:
            _send_capture(
                config=config,
                method="POST",
                url=url,
                request_headers=request_headers,
                request_body=request_body,
                status_code=status_code,
                response_headers=response_headers,
                response_body=response_body,
                duration_s=duration,
            )
        except Exception as capture_err:
            logger.debug("Failed to capture gRPC request: %s", capture_err)

        raise

    try:
        _send_capture(
            config=config,
            method="POST",
            url=url,
            request_headers=request_headers,
            request_body=request_body,
            status_code=status_code,
            response_headers=response_headers,
            response_body=response_body,
            duration_s=duration,
        )
    except Exception as capture_err:
        logger.debug("Failed to capture gRPC request: %s", capture_err)

    return response


def _send_capture(
    config: SmelloConfig,
    method: str,
    url: str,
    request_headers: dict,
    request_body: str,
    status_code: int,
    response_headers: dict,
    response_body: str,
    duration_s: float,
) -> None:
    payload = serialize_request_response(
        config=config,
        method=method,
        url=url,
        request_headers=request_headers,
        request_body=request_body,
        status_code=status_code,
        response_headers=response_headers,
        response_body=response_body,
        duration_s=duration_s,
        library="grpc",
    )
    send(payload)



"""Monkey-patch for the `requests` library."""

import logging
import time
from urllib.parse import urlparse

from smello.capture import serialize_request_response
from smello.config import SmelloConfig
from smello.transport import send

logger = logging.getLogger(__name__)


def patch_requests(config: SmelloConfig) -> None:
    """Patch requests.Session.send to capture outgoing HTTP traffic."""
    try:
        import requests  # noqa: PLC0415 -- optional dependency
    except ImportError:
        return  # requests not installed, skip

    original_send = requests.Session.send

    def patched_send(self, prepared_request, **kwargs):
        host = urlparse(prepared_request.url).hostname or ""

        if not config.should_capture(host):
            return original_send(self, prepared_request, **kwargs)

        start = time.monotonic()
        response = original_send(self, prepared_request, **kwargs)
        duration = time.monotonic() - start

        try:
            payload = serialize_request_response(
                config=config,
                method=prepared_request.method or "GET",
                url=prepared_request.url,
                request_headers=dict(prepared_request.headers),
                request_body=prepared_request.body,
                status_code=response.status_code,
                response_headers=dict(response.headers),
                response_body=response.text,
                duration_s=duration,
                library="requests",
            )
            send(payload)
        except Exception as err:
            logger.debug("Failed to capture request: %s", err)

        return response

    requests.Session.send = patched_send  # type: ignore[assignment]



"""Smello - Capture outgoing HTTP requests automatically."""

import atexit
import logging
from urllib.parse import urlparse

from smello._env import _env_bool, _env_list, _env_str
from smello.config import SmelloConfig
from smello.patches import apply_all as _apply_all
from smello.transport import flush, shutdown
from smello.transport import start_worker as _start_worker

logger = logging.getLogger("smello")
logger.addHandler(logging.NullHandler())

__all__ = ["init", "flush", "shutdown"]
__version__ = "0.5.0"

_DEFAULT_REDACT_HEADERS = ["authorization", "x-api-key"]

_config: SmelloConfig | None = None
_atexit_registered: bool = False


def init(
    server_url: str | None = None,
    capture_hosts: list[str] | None = None,
    capture_all: bool | None = None,
    ignore_hosts: list[str] | None = None,
    redact_headers: list[str] | None = None,
    redact_query_params: list[str] | None = None,
) -> None:
    """Initialize Smello. Patches requests and httpx to capture outgoing HTTP traffic.

    Smello only activates when a server URL is provided — either via the
    ``server_url`` parameter or the ``SMELLO_URL`` environment variable.
    Without a URL, ``init()`` is a no-op: no monkey-patching, no background
    threads, no side effects.  This makes it safe to leave ``smello.init()``
    in production code and control activation purely via the environment.

    Each parameter falls back to a ``SMELLO_*`` environment variable when
    not passed explicitly, then to a hardcoded default:

    ====================  ==============================  ==========================
    Parameter             Environment variable            Default
    ====================  ==============================  ==========================
    server_url            ``SMELLO_URL``                  ``None`` (inactive)
    capture_all           ``SMELLO_CAPTURE_ALL``          ``True``
    capture_hosts         ``SMELLO_CAPTURE_HOSTS``        ``[]``
    ignore_hosts          ``SMELLO_IGNORE_HOSTS``         ``[]``
    redact_headers        ``SMELLO_REDACT_HEADERS``       ``["authorization", "x-api-key"]``
    redact_query_params   ``SMELLO_REDACT_QUERY_PARAMS``  ``[]``
    ====================  ==============================  ==========================

    Boolean env vars accept ``true``/``1``/``yes`` and ``false``/``0``/``no``
    (case-insensitive).  List env vars are comma-separated.
    """
    global _config, _atexit_registered

    # Resolve: explicit param > env var
    if server_url is None:
        server_url = _env_str("URL")
    if not server_url:
        logger.warning(
            "smello.init() called without a server URL. "
            "Set SMELLO_URL or pass server_url= to activate. Doing nothing."
        )
        return

    if capture_all is None:
        env_val = _env_bool("CAPTURE_ALL")
        capture_all = env_val if env_val is not None else True

    if capture_hosts is None:
        capture_hosts = _env_list("CAPTURE_HOSTS") or []

    if ignore_hosts is None:
        ignore_hosts = _env_list("IGNORE_HOSTS") or []

    if redact_headers is None:
        env_headers = _env_list("REDACT_HEADERS")
        redact_headers = (
            env_headers if env_headers is not None else list(_DEFAULT_REDACT_HEADERS)
        )

    if redact_query_params is None:
        redact_query_params = _env_list("REDACT_QUERY_PARAMS") or []

    _config = SmelloConfig(
        server_url=server_url.rstrip("/"),
        capture_hosts=capture_hosts,
        capture_all=capture_all,
        ignore_hosts=ignore_hosts,
        redact_headers=[h.lower() for h in redact_headers],
        redact_query_params=[p.lower() for p in redact_query_params],
    )

    # Always ignore the smello server itself
    server_host = urlparse(_config.server_url).hostname
    if server_host and server_host not in _config.ignore_hosts:
        _config.ignore_hosts.append(server_host)

    # Start transport worker
    _start_worker(_config.server_url)

    # Apply patches
    _apply_all(_config)

    # Register atexit hook once so pending captures are flushed on exit
    if not _atexit_registered:
        atexit.register(shutdown)
        _atexit_registered = True



"""Tests for smello._env helpers and env-var-based init()."""

import logging
import os
from unittest.mock import patch

import pytest
import smello
from smello._env import _env_bool, _env_list, _env_str

# --- _env_str ---


class TestEnvStr:
    def test_returns_value(self):
        with patch.dict(os.environ, {"SMELLO_URL": "http://example.com:9000"}):
            assert _env_str("URL") == "http://example.com:9000"

    def test_strips_whitespace(self):
        with patch.dict(os.environ, {"SMELLO_URL": "  http://host:1234  "}):
            assert _env_str("URL") == "http://host:1234"

    def test_returns_none_when_unset(self):
        with patch.dict(os.environ, {}, clear=True):
            assert _env_str("URL") is None

    def test_returns_none_when_empty(self):
        with patch.dict(os.environ, {"SMELLO_URL": ""}):
            assert _env_str("URL") is None

    def test_returns_none_when_only_whitespace(self):
        with patch.dict(os.environ, {"SMELLO_URL": "   "}):
            assert _env_str("URL") is None


# --- _env_bool ---


class TestEnvBool:
    @pytest.mark.parametrize(
        "value", ["true", "True", "TRUE", "1", "yes", "Yes", "YES"]
    )
    def test_truthy(self, value):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_ALL": value}):
            assert _env_bool("CAPTURE_ALL") is True

    @pytest.mark.parametrize(
        "value", ["false", "False", "FALSE", "0", "no", "No", "NO"]
    )
    def test_falsy(self, value):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_ALL": value}):
            assert _env_bool("CAPTURE_ALL") is False

    def test_returns_none_when_unset(self):
        with patch.dict(os.environ, {}, clear=True):
            assert _env_bool("CAPTURE_ALL") is None

    def test_returns_none_when_empty(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_ALL": ""}):
            assert _env_bool("CAPTURE_ALL") is None

    def test_returns_none_for_unrecognised(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_ALL": "maybe"}):
            assert _env_bool("CAPTURE_ALL") is None


# --- _env_list ---


class TestEnvList:
    def test_comma_separated(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_HOSTS": "a.com,b.com,c.com"}):
            assert _env_list("CAPTURE_HOSTS") == ["a.com", "b.com", "c.com"]

    def test_strips_items(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_HOSTS": " a.com , b.com "}):
            assert _env_list("CAPTURE_HOSTS") == ["a.com", "b.com"]

    def test_skips_empty_items(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_HOSTS": "a.com,,b.com,"}):
            assert _env_list("CAPTURE_HOSTS") == ["a.com", "b.com"]

    def test_returns_none_when_unset(self):
        with patch.dict(os.environ, {}, clear=True):
            assert _env_list("CAPTURE_HOSTS") is None

    def test_returns_none_when_empty(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_HOSTS": ""}):
            assert _env_list("CAPTURE_HOSTS") is None

    def test_single_item(self):
        with patch.dict(os.environ, {"SMELLO_CAPTURE_HOSTS": "api.stripe.com"}):
            assert _env_list("CAPTURE_HOSTS") == ["api.stripe.com"]


# --- init() env var integration ---


class TestInitEnvVars:
    """Test that smello.init() resolves env vars correctly."""

    def test_no_url_does_nothing(self):
        """init() without a URL should not activate."""
        with patch.dict(os.environ, {}, clear=True):
            smello._config = None
            smello.init()
            assert smello._config is None

    def test_no_url_logs_warning(self, caplog):
        """init() without a URL should log a warning."""
        with (
            patch.dict(os.environ, {}, clear=True),
            caplog.at_level(logging.WARNING, logger="smello"),
        ):
            smello._config = None
            smello.init()
            assert "server URL" in caplog.text

    def test_server_url_from_env(self):
        """SMELLO_URL should set the server URL."""

        with (
            patch.dict(os.environ, {"SMELLO_URL": "http://smello:9999"}),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.server_url == "http://smello:9999"

    def test_explicit_server_url_overrides_env(self):
        """Explicit server_url should win over SMELLO_URL."""

        with (
            patch.dict(os.environ, {"SMELLO_URL": "http://from-env:1111"}),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init(server_url="http://explicit:2222")
            assert smello._config.server_url == "http://explicit:2222"

    def test_capture_hosts_from_env(self):
        with (
            patch.dict(
                os.environ,
                {
                    "SMELLO_URL": "http://test:5110",
                    "SMELLO_CAPTURE_HOSTS": "a.com,b.com",
                },
            ),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.capture_hosts == ["a.com", "b.com"]

    def test_ignore_hosts_from_env(self):
        with (
            patch.dict(
                os.environ,
                {
                    "SMELLO_URL": "http://test:5110",
                    "SMELLO_IGNORE_HOSTS": "internal.svc,localhost",
                },
            ),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert "internal.svc" in smello._config.ignore_hosts
            assert "localhost" in smello._config.ignore_hosts

    def test_redact_headers_from_env(self):
        with (
            patch.dict(
                os.environ,
                {
                    "SMELLO_URL": "http://test:5110",
                    "SMELLO_REDACT_HEADERS": "X-Secret,X-Token",
                },
            ),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.redact_headers == ["x-secret", "x-token"]

    def test_redact_query_params_from_env(self):
        with (
            patch.dict(
                os.environ,
                {
                    "SMELLO_URL": "http://test:5110",
                    "SMELLO_REDACT_QUERY_PARAMS": "api_key,token",
                },
            ),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.redact_query_params == ["api_key", "token"]

    def test_redact_query_params_default_empty(self):
        with (
            patch.dict(os.environ, {"SMELLO_URL": "http://test:5110"}, clear=True),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.redact_query_params == []

    def test_redact_headers_default_without_env(self):
        with (
            patch.dict(os.environ, {"SMELLO_URL": "http://test:5110"}, clear=True),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.redact_headers == ["authorization", "x-api-key"]

    def test_capture_all_false_from_env(self):
        with (
            patch.dict(
                os.environ,
                {"SMELLO_URL": "http://test:5110", "SMELLO_CAPTURE_ALL": "false"},
            ),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init()
            assert smello._config.capture_all is False

    def test_defaults_with_explicit_url(self):
        """With only a URL and no other config, hardcoded defaults apply."""

        with (
            patch.dict(os.environ, {}, clear=True),
            patch("smello._start_worker"),
            patch("smello._apply_all"),
        ):
            smello._config = None
            smello.init(server_url="http://test:5110")
            assert smello._config.server_url == "http://test:5110"
            assert smello._config.capture_all is True
            assert smello._config.capture_hosts == []
            assert "test" in smello._config.ignore_hosts  # auto-added
            assert smello._config.redact_headers == ["authorization", "x-api-key"]
            assert smello._config.redact_query_params == []



"""Tests for smello.capture serialization."""

import pytest
from smello.capture import serialize_request_response
from smello.config import SmelloConfig


@pytest.fixture()
def config():
    return SmelloConfig(
        server_url="http://test:5110", redact_headers=["authorization", "x-api-key"]
    )


@pytest.fixture()
def basic_payload(config):
    return serialize_request_response(
        config=config,
        method="GET",
        url="https://api.example.com/test",
        request_headers={"Content-Type": "application/json"},
        request_body=None,
        status_code=200,
        response_headers={"Content-Type": "application/json"},
        response_body='{"ok": true}',
        duration_s=0.15,
        library="requests",
    )


def test_basic_fields(basic_payload):
    assert basic_payload["duration_ms"] == 150
    assert basic_payload["request"]["method"] == "GET"
    assert basic_payload["request"]["url"] == "https://api.example.com/test"
    assert basic_payload["response"]["status_code"] == 200
    assert basic_payload["response"]["body"] == '{"ok": true}'
    assert basic_payload["meta"]["library"] == "requests"
    assert "id" in basic_payload
    assert "timestamp" in basic_payload


def test_null_body(basic_payload):
    assert basic_payload["request"]["body"] is None
    assert basic_payload["request"]["body_size"] == 0


def test_header_redaction(config):
    payload = serialize_request_response(
        config=config,
        method="POST",
        url="https://example.com",
        request_headers={
            "Content-Type": "application/json",
            "Authorization": "Bearer sk-secret",
            "X-Api-Key": "key_12345",
            "X-Custom": "keep-this",
        },
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.1,
        library="httpx",
    )
    headers = payload["request"]["headers"]
    assert headers["Content-Type"] == "application/json"
    assert headers["Authorization"] == "[REDACTED]"
    assert headers["X-Api-Key"] == "[REDACTED]"
    assert headers["X-Custom"] == "keep-this"


def test_custom_redact_headers():
    config = SmelloConfig(server_url="http://test:5110", redact_headers=["x-secret"])
    payload = serialize_request_response(
        config=config,
        method="GET",
        url="https://example.com",
        request_headers={"Authorization": "Bearer token", "X-Secret": "hidden"},
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.05,
        library="requests",
    )
    headers = payload["request"]["headers"]
    assert headers["Authorization"] == "Bearer token"
    assert headers["X-Secret"] == "[REDACTED]"


def test_bytes_body_utf8(config):
    payload = serialize_request_response(
        config=config,
        method="POST",
        url="https://example.com",
        request_headers={},
        request_body=b'{"key": "value"}',
        status_code=200,
        response_headers={},
        response_body=b'{"result": "ok"}',
        duration_s=0.1,
        library="requests",
    )
    assert payload["request"]["body"] == '{"key": "value"}'
    assert payload["request"]["body_size"] == 16
    assert payload["response"]["body"] == '{"result": "ok"}'


def test_binary_body_non_utf8(config):
    binary_data = bytes(range(256))
    payload = serialize_request_response(
        config=config,
        method="POST",
        url="https://example.com",
        request_headers={},
        request_body=binary_data,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.1,
        library="requests",
    )
    assert payload["request"]["body"] == "<binary: 256 bytes>"
    assert payload["request"]["body_size"] == 256


def test_string_body(config):
    payload = serialize_request_response(
        config=config,
        method="POST",
        url="https://example.com",
        request_headers={},
        request_body="plain text",
        status_code=200,
        response_headers={},
        response_body="response text",
        duration_s=0.1,
        library="httpx",
    )
    assert payload["request"]["body"] == "plain text"
    assert payload["response"]["body"] == "response text"


def test_query_param_redaction():
    config = SmelloConfig(
        server_url="http://test:5110",
        redact_query_params=["api_key", "token"],
    )
    payload = serialize_request_response(
        config=config,
        method="GET",
        url="https://example.com/search?q=hello&api_key=secret123&token=abc&page=1",
        request_headers={},
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.1,
        library="requests",
    )
    url = payload["request"]["url"]
    assert "q=hello" in url
    assert "page=1" in url
    assert "secret123" not in url
    assert "abc" not in url
    assert "api_key=%5BREDACTED%5D" in url
    assert "token=%5BREDACTED%5D" in url


def test_query_param_redaction_case_insensitive():
    config = SmelloConfig(
        server_url="http://test:5110",
        redact_query_params=["api_key"],
    )
    payload = serialize_request_response(
        config=config,
        method="GET",
        url="https://example.com/get?API_KEY=secret&other=keep",
        request_headers={},
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.1,
        library="requests",
    )
    url = payload["request"]["url"]
    assert "secret" not in url
    assert "other=keep" in url


def test_query_param_redaction_no_query_string():
    config = SmelloConfig(
        server_url="http://test:5110",
        redact_query_params=["api_key"],
    )
    payload = serialize_request_response(
        config=config,
        method="GET",
        url="https://example.com/path",
        request_headers={},
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.1,
        library="requests",
    )
    assert payload["request"]["url"] == "https://example.com/path"


def test_query_param_redaction_empty_list():
    """No redact_query_params means URL is left untouched."""
    config = SmelloConfig(server_url="http://test:5110")
    payload = serialize_request_response(
        config=config,
        method="GET",
        url="https://example.com/get?api_key=secret",
        request_headers={},
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=0.1,
        library="requests",
    )
    assert payload["request"]["url"] == "https://example.com/get?api_key=secret"


def test_duration_rounding(config):
    payload = serialize_request_response(
        config=config,
        method="GET",
        url="https://example.com",
        request_headers={},
        request_body=None,
        status_code=200,
        response_headers={},
        response_body=None,
        duration_s=1.5678,
        library="requests",
    )
    assert payload["duration_ms"] == 1567



"""Serialize captured HTTP request/response pairs for sending to the server."""

import sys
import time
import uuid
from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit

from smello.config import SmelloConfig


def serialize_request_response(
    config: SmelloConfig,
    method: str,
    url: str,
    request_headers: dict,
    request_body: str | bytes | None,
    status_code: int,
    response_headers: dict,
    response_body: str | bytes | None,
    duration_s: float,
    library: str,
) -> dict:
    """Build the capture payload dict."""
    req_headers = _redact_headers(dict(request_headers), config.redact_headers)
    url = _redact_query_params(url, config.redact_query_params)
    resp_headers = dict(response_headers)

    req_body_str = _body_to_str(request_body)
    resp_body_str = _body_to_str(response_body)

    return {
        "id": str(uuid.uuid4()),
        "timestamp": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        "duration_ms": int(duration_s * 1000),
        "request": {
            "method": method,
            "url": url,
            "headers": req_headers,
            "body": req_body_str,
            "body_size": len(request_body) if request_body else 0,
        },
        "response": {
            "status_code": status_code,
            "headers": resp_headers,
            "body": resp_body_str,
            "body_size": len(response_body) if response_body else 0,
        },
        "meta": {
            "library": library,
            "python_version": _python_version(),
            "smello_version": "0.1.0",
        },
    }


def _redact_headers(headers: dict, redact_keys: list[str]) -> dict:
    return {
        k: ("[REDACTED]" if k.lower() in redact_keys else v) for k, v in headers.items()
    }


def _redact_query_params(url: str, redact_keys: list[str]) -> str:
    if not redact_keys:
        return url
    parts = urlsplit(url)
    if not parts.query:
        return url
    params = parse_qs(parts.query, keep_blank_values=True)
    redacted = {
        k: (["[REDACTED]"] * len(v) if k.lower() in redact_keys else v)
        for k, v in params.items()
    }
    new_query = urlencode(redacted, doseq=True)
    return urlunsplit(parts._replace(query=new_query))


def _body_to_str(body: str | bytes | None) -> str | None:
    if body is None:
        return None
    if isinstance(body, bytes):
        try:
            return body.decode("utf-8")
        except UnicodeDecodeError:
            return f"<binary: {len(body)} bytes>"
    return body


def _python_version() -> str:
    return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
